// Copyright (c) 2009, 2010 Paul Pogonyshev.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.


#include "tests/common.hpp"

#include <map>

#include <boost/mpl/end.hpp>
#include <boost/mpl/insert_range.hpp>
#include <boost/mpl/push_back.hpp>
#include <boost/mpl/vector.hpp>
#include <boost/test/unit_test.hpp>

#if HAVE_UNORDERED_SET
#  include <unordered_set>
using std::unordered_set;

#elif HAVE_TR1_UNORDERED_SET
#  include <tr1/unordered_set>
using std::tr1::unordered_set;

#elif HAVE_BOOST_UNORDERED_SET_HPP
#  include <boost/unordered_set.hpp>
using boost::unordered_set;
#endif


namespace mct
{

#define TEST_SET_OUTPUT(tested_set_type)                                                \
  template <typename Value, typename Hash, typename Equal, typename Allocator,          \
            bool keep_hashes>                                                           \
  ostream&                                                                              \
  operator<< (ostream& stream,                                                          \
              const tested_set_type <Value, Hash, Equal, Allocator, keep_hashes>& set)  \
  {                                                                                     \
    return print_range (stream, "{}", set.begin (), set.end ());                        \
  }

  TEST_SET_OUTPUT (closed_hash_set);
  TEST_SET_OUTPUT (linked_hash_set);

}


namespace std
{

  template <typename Value, typename Hash, typename Equal, typename Allocator>
  ostream&
  operator<< (ostream& stream, const unordered_set <Value, Hash, Equal, Allocator>& set)
  {
    return print_range (stream, "{}", set.begin (), set.end ());
  }

}


namespace
{

#define TEST_SET_COMPARISON(tested_set_type)                                                    \
  template <typename Value, typename Hash, typename Equal, typename Allocator,                  \
            bool keep_hashes>                                                                   \
  bool                                                                                          \
  operator== (const tested_set_type <Value, Hash, Equal, leak_test_allocator <Allocator>,       \
                                     keep_hashes>& tested_set,                                  \
              const unordered_set <Value, Hash, Equal, Allocator>& robust_set)                  \
  {                                                                                             \
    if (tested_set.size () != robust_set.size ())                                               \
      return false;                                                                             \
    foreach (const Value& value, robust_set)                                                    \
      {                                                                                         \
        if (!contains (tested_set, value))                                                      \
          return false;                                                                         \
      }                                                                                         \
    return true;                                                                                \
  }

  TEST_SET_COMPARISON (closed_hash_set);
  TEST_SET_COMPARISON (linked_hash_set);


  template <typename base_type>
  class test_set : public base_type
  {
  public:

    typedef  typename base_type::key_type        key_type;
    typedef  typename base_type::value_type      value_type;
    typedef  typename base_type::iterator        iterator;
    typedef  typename base_type::const_iterator  const_iterator;
    typedef  typename base_type::hasher          hasher;
    typedef  typename base_type::key_equal       key_equal;
    typedef  typename base_type::allocator_type  allocator_type;
    typedef  typename allocator_type::base_type  allocator_base;

  private:

    // Note that we don't use leak test allocator for the set we check against.
    typedef  unordered_set <key_type, hasher, key_equal, allocator_base>  robust_type;


    robust_type  _valid;


  public:

    explicit
    test_set (size_t                num_buckets = 0,
              const hasher&         hash        = hasher (),
              const key_equal&      equal       = key_equal (),
              const allocator_type& allocator   = allocator_type ())
      : base_type (num_buckets, hash, equal, allocator),
        _valid    (num_buckets, hash, equal, allocator)
    {
      validate ();
    }

    template <typename InputIterator>
    test_set (InputIterator begin, InputIterator end,
              size_t                num_buckets = 0,
              const hasher&         hash        = hasher (),
              const key_equal&      equal       = key_equal (),
              const allocator_type& allocator   = allocator_type ())
      : base_type (begin, end, num_buckets, hash, equal, allocator),
        _valid    (begin, end, num_buckets, hash, equal, allocator)
    {
      validate ();
    }

    ~test_set ()
    {
      validate ();
    }


    pair <iterator, bool>
    insert (const value_type& value)
    {
      const pair <iterator, bool>  result (base_type::insert (value));
      _valid.insert (value);
      validate ();
      return result;
    }

    iterator
    insert (const_iterator hint, const value_type& value)
    {
      const iterator  result (base_type::insert (hint, value));
      _valid.insert (_valid.end (), value);
      validate ();
      return result;
    }

    template <typename InputIterator>
    void
    insert (InputIterator begin, InputIterator end)
    {
      base_type::insert (begin, end);
      _valid.insert (begin, end);
      validate ();
    }

    bool
    erase (const key_type& key)
    {
      const bool  result = base_type::erase (key);
      BOOST_REQUIRE_EQUAL (result, _valid.erase (key));
      validate ();
      return result;
    }

    iterator
    erase (const_iterator position)
    {
      _valid.erase (*position);
      const iterator  result = base_type::erase (position);
      validate ();
      return result;
    }

    iterator
    erase (const_iterator first, const_iterator last)
    {
      // Order is not required to be the same; emulate.
      for (const_iterator scan = first; scan != last; ++scan)
        _valid.erase (*scan);

      const iterator  result = base_type::erase (first, last);
      validate ();
      return result;
    }


    void
    clear ()
    {
      base_type::clear ();
      _valid.clear ();
      validate ();
    }


    void
    swap (test_set& that)
    {
      base_type::swap (that);
      _valid.swap (that._valid);
      validate ();
      that.validate ();
    }


    void
    rehash (size_t num_buckets)
    {
      base_type::rehash (num_buckets);
      validate ();
    }


    void
    validate () const
    {
      base_type::validate_integrity ();
      BOOST_REQUIRE_EQUAL (*this, _valid);
    }
  };


  template <typename base_type>
  inline  void
  swap (test_set <base_type>& set1, test_set <base_type>& set2)
  {
    set1.swap (set2);
  }


  struct Plain
  {
    template <typename Type, typename Hash, typename Equal, bool keep_hashes>
    struct resolve
    {
      typedef  closed_hash_set <Type, Hash, Equal,
                                leak_test_allocator <allocator <Type> >,
                                keep_hashes>
               set_type;
    };

    static  const bool  linked = false;

    // Hide difference in available methods: for non-linked sets 'before' is just ignored.
    template <typename Set>
    static  pair <typename Set::iterator, bool>
    insert_before (Set& set, typename Set::const_iterator before,
                   const typename Set::value_type& value)
    {
      return set.insert (value);
    }
  };

  struct Linked
  {
    template <typename Type, typename Hash, typename Equal, bool keep_hashes>
    struct resolve
    {
      typedef  linked_hash_set <Type, Hash, Equal,
                                leak_test_allocator <allocator <Type> >,
                                keep_hashes>
               set_type;
    };

    static  const bool  linked = true;

    template <typename Set>
    static  pair <typename Set::iterator, bool>
    insert_before (Set& set, typename Set::const_iterator before,
                   const typename Set::value_type& value)
    {
      return set.insert (before, value);
    }
  };


  template <typename Implementation,
            typename Type,
            typename Hash  = MCT_HASH_NAMESPACE::hash <Type>,
            typename Equal = equal_to <Type>,
            bool _keep_hashes = false>
  struct Parameters
  {
    typedef  Implementation  implementation;
    typedef  Type            type;
    typedef  Hash            hasher;
    typedef  Equal           equal_to;

    static  const bool  keep_hashes = _keep_hashes;
    static  const bool  linked      = Implementation::linked;

    typedef  test_set <typename Implementation::
                       template resolve <Type, Hash, Equal, _keep_hashes>::set_type>
             set_type;
  };


  template <typename Implementation, typename Type, bool _keep_hashes = false>
  struct ExceptionSafetyParameters
  {
    typedef  Implementation  implementation;
    typedef  Type            type;

    static  const bool  keep_hashes = _keep_hashes;
    static  const bool  linked      = Implementation::linked;

    // Not 'test_set' here as we can't control where exception are actually thrown.
    typedef  typename Implementation::template resolve <Type,
                                                        MCT_HASH_NAMESPACE::hash <type>,
                                                        equal_to <type>,
                                                        keep_hashes>::set_type
             set_type;
  };


  template <typename type>
  vector <type>
  create_vector (type (* generator) (int), int stop, int start = 0)
  {
    vector <type>  result;

    if (stop > start)
      {
        result.reserve (stop - start);
        for (int k = start; k < stop; ++k)
          result.push_back (generator (k));
      }

    return result;
  }

  template <typename Set, typename Initializer>
  void
  assert_identical_order (const Set& set, const Initializer& initializer)
  {
    typedef  typename Set::value_type  value_type;
    typedef  map <int, value_type>     order_map;

    vector <value_type>    actual_order;
    vector <value_type>    expected_order;
    order_map              order;

    foreach (const value_type& value, set)
      {
        actual_order.push_back (value);

        int  index = 0;
        foreach (const value_type& candidate, initializer)
          {
            if (set.key_eq () (value, candidate))
              {
                order.insert (make_pair (index, value));
                break;
              }

            ++index;
          }
      }

    foreach (const typename order_map::value_type& entry, order)
      expected_order.push_back (entry.second);

    BOOST_CHECK_EQUAL (actual_order, expected_order);
  }


  // To speed compilation up we list just a few possible combinations for each base type.
  typedef  mpl::vector <Parameters <Plain, int>,
                        Parameters <Plain, int, int_hasher_1>,
                        Parameters <Plain, int,
                                    MCT_HASH_NAMESPACE::hash <int>, equal_to <int>, true>,
#                     if MCT_HAVE_LONG_LONG
                        Parameters <Plain, unsigned long long>,
                        Parameters <Plain, unsigned long long, unsigned_long_long_hasher>,
#                     endif
                        Parameters <Plain, string>,
                        Parameters <Plain, string, string_hasher_1>,
                        Parameters <Plain, string, string_hasher_1, equal_to <string>, true>,
                        Parameters <Plain, foo> >
           test_parameters_closed_std_equal;

  typedef  mpl::vector <Parameters <Linked, int>,
#                     if MCT_HAVE_LONG_LONG
                        Parameters <Linked, unsigned long long, unsigned_long_long_hasher,
                                    equal_to <unsigned long long>, true>,
#                     endif
                        Parameters <Linked, string, string_hasher_1>,
                        Parameters <Linked, foo> >
           test_parameters_linked_std_equal;

  typedef  test_parameters_linked_std_equal
           test_parameters_linked;

  typedef  mpl::insert_range <test_parameters_closed_std_equal,
                              mpl::end <test_parameters_closed_std_equal>::type,
                              test_parameters_linked_std_equal>::type
           test_parameters_std_equal;

  typedef  mpl::push_back <test_parameters_std_equal,
                           Parameters <Plain, int, int_hasher_special,
                                       int_comparator_special> >::type
           test_parameters;


#define COMMON_TEST_SETUP                                       \
  typedef  typename parameters::implementation  implementation; \
  typedef  typename parameters::set_type        set_type;       \
  typedef  typename set_type::iterator          iterator;       \
  typedef  typename set_type::const_iterator    const_iterator; \
  typedef  typename parameters::type            type;           \
  typedef  typename parameters::hasher          hasher;         \
  typedef  typename parameters::equal_to        equal_to;       \
  typedef  Data <type>                          data;           \
  const bool  linked = parameters::linked


  typedef  mpl::vector <ExceptionSafetyParameters <Plain,  throws_on_copy, false>,
                        ExceptionSafetyParameters <Plain,  throws_on_copy, true>,
                        ExceptionSafetyParameters <Plain,  throws_in_comparator, false>,
                        ExceptionSafetyParameters <Plain,  throws_in_comparator, true>,
                        ExceptionSafetyParameters <Plain,  throws_in_hasher, false>,
                        ExceptionSafetyParameters <Plain,  throws_in_hasher, true>,
                        ExceptionSafetyParameters <Linked, throws_on_copy, false>,
                        ExceptionSafetyParameters <Linked, throws_on_copy, true>,
                        ExceptionSafetyParameters <Linked, throws_in_comparator, false>,
                        ExceptionSafetyParameters <Linked, throws_in_comparator, true>,
                        ExceptionSafetyParameters <Linked, throws_in_hasher, false>,
                        ExceptionSafetyParameters <Linked, throws_in_hasher, true> >
           exception_safety_test_parameters;


#define COMMON_EXCEPTION_SAFETY_TEST_SETUP                              \
  typedef  typename parameters::implementation  implementation;         \
  typedef  typename parameters::set_type        set_type;               \
  typedef  typename set_type::iterator          iterator;               \
  typedef  typename set_type::const_iterator    const_iterator;         \
  typedef  typename parameters::type            type;                   \
  typedef  typename type::enabler               exception_enabler;      \
  const bool  linked = parameters::linked


# include "tests/common-data.hpp"

}


BOOST_AUTO_TEST_SUITE (set)


// Note that the tests do not look like checking almost anything.  That's because
// 'test_set' validates and compares itself with a known good implementation after each
// modification.  So, most actual checks just happen elsewhere.

BOOST_AUTO_TEST_CASE_TEMPLATE (test_range_constructor_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (ARRAY_RANGE (data::values1));
  if (linked)
    assert_identical_order (set, data::values1);
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_find_1,
                               parameters, test_parameters_std_equal)
{
  COMMON_TEST_SETUP;

  set_type  set (ARRAY_RANGE (data::values1));

  foreach (const type& value, data::values1)
    BOOST_CHECK (contains (set, value));

  foreach (const type& value, data::values2)
    {
      BOOST_CHECK_EQUAL (contains (set, value),
                         find (ARRAY_RANGE (data::values1), value) != ARRAY_END (data::values1));
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_count_1,
                               parameters, test_parameters_std_equal)
{
  COMMON_TEST_SETUP;

  set_type  set (ARRAY_RANGE (data::values1));

  foreach (const type& value, data::values1)
    BOOST_CHECK_EQUAL (set.count (value), 1u);

  foreach (const type& value, data::values2)
    {
      BOOST_CHECK_EQUAL (set.count (value),
                         static_cast <size_t> (count (ARRAY_RANGE (data::values1), value)));
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_equal_range_1,
                               parameters, test_parameters_std_equal)
{
  COMMON_TEST_SETUP;

  set_type  set (ARRAY_RANGE (data::values1));

  foreach (const type& value, data::values1)
    {
      pair <const_iterator, const_iterator>  range (set.equal_range (value));
      BOOST_CHECK_EQUAL (*range.first, value);
      BOOST_CHECK_EQUAL (++range.first, range.second);
    }

  foreach (const type& value, data::values2)
    {
      pair <const_iterator, const_iterator>  range (set.equal_range (value));

      if (find (ARRAY_RANGE (data::values1), value) == ARRAY_END (data::values1))
        BOOST_CHECK_EQUAL (range.first, range.second);
      else
        {
          BOOST_CHECK_EQUAL (*range.first, value);
          BOOST_CHECK_EQUAL (++range.first, range.second);
        }
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_comparison_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (ARRAY_RANGE (data::values1));
  set_type  set2 (ARRAY_RANGE (data::values1));
  set_type  set3 (ARRAY_RANGE (data::values2));
  set_type  set4 (ARRAY_RANGE (data::values2));

  BOOST_CHECK_EQUAL (set1, set2);
  BOOST_CHECK_EQUAL (set3, set4);
  BOOST_CHECK_NE    (set1, set3);
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_copy_constructor_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (ARRAY_RANGE (data::values1));
  set_type  set2 (ARRAY_RANGE (data::values2));
  set_type  set3 (set1);
  set_type  set4 (set2);

  BOOST_CHECK_EQUAL (set3, set1);
  BOOST_CHECK_EQUAL (set4, set2);

  if (linked)
    {
      assert_identical_order (set3, set1);
      assert_identical_order (set4, set2);
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_assignment_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (ARRAY_RANGE (data::values1));
  set_type  set2;

  BOOST_CHECK_EQUAL (set2.size (), 0u);

  set2 = set1;
  BOOST_CHECK_EQUAL (set2, set1);

  if (linked)
    assert_identical_order (set1, set2);
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_self_assignment_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (ARRAY_RANGE (data::values1));
  set_type  set2 (ARRAY_RANGE (data::values1));

  BOOST_CHECK_EQUAL (set1, set2);
  if (linked)
    assert_identical_order (set1, set2);

  set1 = set1;

  BOOST_CHECK_EQUAL (set1, set2);
  if (linked)
    assert_identical_order (set1, set2);
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_iterators_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type        set (ARRAY_RANGE (data::values1));
  size_t          size = set.size ();
  const_iterator  scan = set.begin ();

  for (size_t k = 0; k != size; ++k, ++scan)
    {
      BOOST_CHECK (contains (set, *scan));

      if (k != 0)
        {
          const_iterator  copy (scan);
          --copy;
          ++copy;
          BOOST_CHECK_EQUAL (copy, scan);

          const_iterator  copy2 (scan);
          copy2--;
          copy2++;
          BOOST_CHECK_EQUAL (copy2, scan);
        }

      const_iterator  copy (scan);
      ++copy;
      --copy;
      BOOST_CHECK_EQUAL (copy, scan);

      const_iterator  copy2 (scan);
      copy2++;
      copy2--;
      BOOST_CHECK_EQUAL (copy2, scan);
    }

  BOOST_CHECK_EQUAL (scan, set.end ());
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_swap_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (ARRAY_RANGE (data::values1));
  set_type  set2 (ARRAY_RANGE (data::values2));
  set_type  set3 (ARRAY_RANGE (data::values1));
  set_type  set4 (ARRAY_RANGE (data::values2));

  BOOST_CHECK_EQUAL (set1, set3);
  BOOST_CHECK_EQUAL (set2, set4);

  if (linked)
    {
      assert_identical_order (set1, set3);
      assert_identical_order (set2, set4);
    }

  swap (set3, set4);

  BOOST_CHECK_EQUAL (set1, set4);
  BOOST_CHECK_EQUAL (set2, set3);

  if (linked)
    {
      assert_identical_order (set1, set4);
      assert_identical_order (set2, set3);
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_clear_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (ARRAY_RANGE (data::values1));
  set.clear ();
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_erase_value_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (ARRAY_RANGE (data::values1));
  set_type  set2 (ARRAY_BEGIN (data::values1) + 1, ARRAY_END (data::values1));

  // Make sure funny comparison functions don't ruin this test.
  if (set1 != set2)
    {
      BOOST_CHECK       (set1.erase (data::values1[0]));
      BOOST_CHECK_EQUAL (set1, set2);

      if (linked)
        assert_identical_order (set1, set2);
    }
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_erase_position_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (ARRAY_RANGE (data::values1));

  const size_t  num_passes_expected = set.size ();
  size_t        num_passes          = 0;

  for (iterator scan = set.begin (), end = set.end (); scan != end; ++num_passes)
    scan = set.erase (scan);

  BOOST_CHECK       (set.empty ());
  BOOST_CHECK_EQUAL (num_passes, num_passes_expected);
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_erase_range_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (ARRAY_RANGE (data::values1));

  set.erase (set.begin (), set.end ());
  BOOST_CHECK (set.empty ());
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_growth_1,
                               parameters, test_parameters_std_equal)
{
  COMMON_TEST_SETUP;

  // Note: fairly small, because testing after each step is slow.
  set_type  set;
  for (int k = 0; k < 1000; ++k)
    set.insert (data::generate (k));

  BOOST_CHECK_EQUAL (set.size (), 1000u);
  if (linked)
    assert_identical_order (set, create_vector (data::generate, set.size ()));
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_shrink_1,
                               parameters, test_parameters_std_equal)
{
  COMMON_TEST_SETUP;

  vector <type>  items (create_vector (data::generate, 10000));
  set_type       set   (items.begin (), items.end ());

  BOOST_CHECK_EQUAL (set.size (), 10000u);

  const_iterator  almost_end = set.end ();
  for (int k = 0; k < 10; ++k)
    --almost_end;

  set.erase (set.begin (), almost_end);

  BOOST_CHECK_EQUAL (set.size (), 10u);
  if (linked)
    assert_identical_order (set, create_vector (data::generate, 10000, 10000 - 10));

  set.rehash (32);

  BOOST_CHECK_EQUAL (set.bucket_count (), 32u);
  if (linked)
    assert_identical_order (set, create_vector (data::generate, 10000, 10000 - 10));
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_tiny_set_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  float  load_factors[] = { 0.01, 0.1, 0.25, 0.5, 0.75, 0.9, 0.99 };

  foreach (float load_factor, load_factors)
    {
      BOOST_TEST_CHECKPOINT ("testing with maximum load factor " << load_factor);

      set_type  set (1);
      set.max_load_factor (load_factor);

      for (int k = 0; k < 50; ++k)
        set.insert (data::generate (k));

      if (linked)
        assert_identical_order (set, create_vector (data::generate, 50));
    }
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_relink_1,
                               parameters, test_parameters_linked)
{
  COMMON_TEST_SETUP;

  vector <type>  initializers (create_vector (data::generate, 50));
  set_type       set (initializers.begin (), initializers.end ());

  assert_identical_order (set, initializers);

  set.relink (set.find (data::generate (40)), set.find (data::generate (20)));
  set.validate_integrity ();

  initializers.erase  (find (initializers.begin (), initializers.end (), data::generate (20)));
  initializers.insert (find (initializers.begin (), initializers.end (), data::generate (40)),
                       data::generate (20));

  assert_identical_order (set, initializers);

  // The following "moves" have no effect.

  set.relink (set.find (data::generate (5)), set.find (data::generate (5)));
  set.validate_integrity ();

  assert_identical_order (set, initializers);

  set.relink (set.find (data::generate (15)), set.find (data::generate (14)));
  set.validate_integrity ();

  assert_identical_order (set, initializers);
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_reverse_1,
                               parameters, test_parameters_linked)
{
  COMMON_TEST_SETUP;

  vector <type>  initializers (create_vector (data::generate, 50));
  set_type       set (initializers.begin (), initializers.end ());

  assert_identical_order (set, initializers);

  set.reverse ();
  set.validate_integrity ();
  reverse (initializers.begin (), initializers.end ());

  assert_identical_order (set, initializers);
}


// For constructors exception safety is practically limited to 'no leaks'.
BOOST_AUTO_TEST_CASE_TEMPLATE (test_range_constructor_exception_safety_1,
                               parameters, exception_safety_test_parameters)
{
  COMMON_EXCEPTION_SAFETY_TEST_SETUP;

  type  initializers[50];
  for (size_t k = 0; k < ARRAY_SIZE (initializers); ++k)
    initializers[k] = type (k);

  for (int fail_after = 0; fail_after < 100; ++fail_after)
    {
      BOOST_TEST_CHECKPOINT ("fail after " << fail_after << " operations");

      {
        // The exception will probably never get thrown for large 'fail_after' values.
        exception_enabler  _(fail_after);

        try
          {
            set_type (ARRAY_RANGE (initializers));
          }
        catch (expected_exception&)
          { }
      }
    }
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_copy_constructor_exception_safety_1,
                               parameters, exception_safety_test_parameters)
{
  COMMON_EXCEPTION_SAFETY_TEST_SETUP;

  type  initializers[50];
  for (size_t k = 0; k < ARRAY_SIZE (initializers); ++k)
    initializers[k] = type (k);

  for (int fail_after = 0; fail_after < 100; ++fail_after)
    {
      BOOST_TEST_CHECKPOINT ("fail after " << fail_after << " operations");

      // Recreate base set each iteration, because our allocator only checks for "no
      // leaks" when all containers using it are destroyed.
      set_type  set (ARRAY_RANGE (initializers));

      {
        // The exception will probably never get thrown for large 'fail_after' values.
        exception_enabler  _(fail_after);

        try
          {
            set_type  copy (set);
          }
        catch (expected_exception&)
          { }
      }
    }
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_exception_safety_1,
                               parameters, exception_safety_test_parameters)
{
  COMMON_EXCEPTION_SAFETY_TEST_SETUP;

  for (int fail_after = 0; fail_after < 100; ++fail_after)
    {
      BOOST_TEST_CHECKPOINT ("fail after " << fail_after << " operations");

      set_type  set;
      size_t    num_inserted = 0;

      {
        exception_enabler  _(fail_after);

        try
          {
            // Limit set size since not all types will eventually throw.
            for (; num_inserted < 10000; ++num_inserted)
              set.insert (type (num_inserted));
          }
        catch (expected_exception&)
          { }
      }

      set.validate_integrity ();
      BOOST_CHECK_EQUAL (set.size (), num_inserted);
    }
}

// Here we test "reinserts", which basically results in inserting over debris buckets.
BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_exception_safety_2,
                               parameters, exception_safety_test_parameters)
{
  COMMON_EXCEPTION_SAFETY_TEST_SETUP;

  for (int fail_after = 0; fail_after < 100; ++fail_after)
    {
      BOOST_TEST_CHECKPOINT ("fail after " << fail_after << " operations");

      set_type  set;
      for (int k = 0; k < 100; ++k)
        set.insert (type (k));

      bool  thrown = false;

      {
        exception_enabler  _(fail_after);

        try
          {
            for (const_iterator scan = set.begin (); scan != set.end ();)
              {
                // Don't copy the type itself, as this may throw.
                const int  value = scan->value;
                scan = set.erase (scan);
                implementation::insert_before (set, scan, type (value));

                // Properly functioning set will not rehash itself in this loop, so
                // iterators must not be invalidated midway.
                BOOST_CHECK (scan == set.end () || set.valid_iterator (scan));
              }
          }
        catch (expected_exception&)
          {
            thrown = true;
          }
      }

      set.validate_integrity ();
      BOOST_CHECK_EQUAL (set.size (), 100u - (thrown ? 1 : 0));
    }
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_range_exception_safety_1,
                               parameters, exception_safety_test_parameters)
{
  COMMON_EXCEPTION_SAFETY_TEST_SETUP;

  type  initializers[100];
  for (size_t k = 0; k < ARRAY_SIZE (initializers); ++k)
    initializers[k] = type (k);

  for (int fail_after = 0; fail_after < 100; ++fail_after)
    {
      BOOST_TEST_CHECKPOINT ("fail after " << fail_after << " operations");

      // Some elements are there from the start, some will be inserted.
      size_t    num_preexisted_elements = (ARRAY_SIZE (initializers) / 2);
      set_type  set (initializers, initializers + num_preexisted_elements);
      bool      thrown = false;

      {
        exception_enabler  _(fail_after);

        try
          {
            set.insert (initializers + num_preexisted_elements, ARRAY_END (initializers));
          }
        catch (expected_exception&)
          {
            thrown = true;
          }
      }

      set.validate_integrity ();

      if (!thrown)
        BOOST_CHECK_EQUAL (set.size (), ARRAY_SIZE (initializers));
      else if (!linked)
        {
          // Only basic exception safety, so the set needn't contain any specific number
          // of elements.
          BOOST_CHECK_GE (set.size (), num_preexisted_elements);
          BOOST_CHECK_LE (set.size (), ARRAY_SIZE (initializers));
        }
      else
        {
          // Linked tables guarantee strong exception safety for this method.
          BOOST_CHECK_EQUAL (set.size (), num_preexisted_elements);
        }

      for (size_t k = 0; k < set.size (); ++k)
        BOOST_CHECK (contains (set, type (k)));
    }
}


BOOST_AUTO_TEST_SUITE_END ()


// Local variables:
// mode: c++
// c-basic-offset: 2
// indent-tabs-mode: nil
// fill-column: 90
// End:
